package com.miya.draw.manager;

import cn.hutool.core.bean.copier.BeanCopier;
import cn.hutool.core.bean.copier.CopyOptions;
import cn.hutool.core.date.LocalDateTimeUtil;
import com.miya.draw.config.constant.ConStants;
import com.miya.draw.config.exception.BusinessException;
import com.miya.draw.entity.*;
import com.miya.draw.enums.*;
import com.miya.draw.model.DrawModel;
import com.miya.draw.request.DrawUserInfoGainRequest;
import com.miya.draw.service.*;
import com.miya.draw.service.impl.ActivityConfigServiceImpl;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * 抽奖Manager
 *
 * @author jack
 */
@Component
@Slf4j
public class DrawManagerImpl implements IDrawManager {

    @Autowired
    private IAccountService accountService;


    @Autowired
    private IAccountPrizeService accountPrizeService;

    @Autowired
    private ISyncTradeService syncTradeService;

    @Autowired
    private IActivityStoreService activityStoreService;

    @Autowired
    private IActivityGoodsService activityGoodsService;

    @Autowired
    private RedissonClient redissonClient;

    @Autowired
    private IPrizeService prizeService;

    @Autowired
    private IOtherManager otherManager;

    @Autowired
    private IActivityConfigService activityConfigService;

    @Override
    public DrawModel gainActivityUserInfo(DrawUserInfoGainRequest request) {
        try {
            Account account = new Account();
            account.setName(StringUtils.isNotBlank(request.getUserName()) ? request.getUserName() : request.getPhone());
            account.setPhone(request.getPhone());
            account.setToken(request.getToken());
            accountService.saveOrUpdateNotById(account);
        } catch (Exception e) {
            log.error("保存用户信息出错:{}", e.getMessage());
        }
        ActivityConfig activityConfig = activityConfigService.getConfig();
        if(activityConfig == null){
            throw  new BusinessException(ErrorEnum.SYSTEM_EXP, "未获取活动有效配置");
        }
        LocalDateTime nowTime = LocalDateTime.now();
        LocalDateTime start = nowTime.minusDays(activityConfig.getValidTime()-1);
        start = LocalDateTimeUtil.beginOfDay(start);
        LocalDateTime end = LocalDateTimeUtil.endOfDay(nowTime);
        Integer maxDrawCountLimit = activityConfig.getDrawMaxCount();
        Integer alreadyDrawCount = accountPrizeService.gainAlreadyDrawCount(request.getPhone(), activityConfig.getActivityId(), start, end);
        List<String> storeIdList = activityStoreService.listAllActivityStoreId(activityConfig.getActivityId());
        List<String> goodCodeList = activityGoodsService.listAllActivityGoods(activityConfig.getActivityId());
        List<SyncTrade> buyGoodList = syncTradeService.listCanDrawGoods(request.getPhone(), start, end, storeIdList, goodCodeList);
        BigDecimal buyTotal = gainBuyTotal(buyGoodList);
        Long purchasedAmount = buyTotal.multiply(BigDecimal.valueOf(100)).longValue();
        Integer drawCountTotal =0;
        if(activityConfig.getRaffle()== RaffleEnum.AMOUNTUNIT.getCode()){
             buyTotal = getAmountTotal(buyGoodList);
             drawCountTotal = buyTotal.divide(new BigDecimal(activityConfig.getAmountUnit()), RoundingMode.DOWN).intValue();
            //足量抽奖
        }else{
            drawCountTotal = buyTotal.divide(new BigDecimal(activityConfig.getCountUnit()), RoundingMode.DOWN).intValue();

        }
        drawCountTotal = Math.min(activityConfig.getDrawMaxCount(), drawCountTotal);
        drawCountTotal = drawCountTotal - alreadyDrawCount;
        Integer joinActivityPeopleNum = accountPrizeService.countJoinActivityPeople(activityConfig.getActivityId());
        return DrawModel.builder()
                .alreadyDrawCount(alreadyDrawCount)
                .drawCountTotal(drawCountTotal>=0?drawCountTotal:0)
                .maxDrawCountLimit(maxDrawCountLimit)
                .numOfDrawCount(Math.max(0, maxDrawCountLimit - alreadyDrawCount))
                .purchasedAmount(purchasedAmount)
                .taxAmountCache(Math.max(0L, BigDecimal.valueOf(activityConfig.getMaxMoney()).subtract(buyTotal).multiply(BigDecimal.valueOf(100)).longValue()))
                .numOfJoinActivityPeople(joinActivityPeopleNum)
                .build();
    }

    /**
     * 获取购买的交易总额
     *
     * @param buyGoodList 购买商品列表
     * @return 购买总金额
     */
    private BigDecimal gainBuyTotal(List<SyncTrade> buyGoodList) {
        BigDecimal buyTotal = buyGoodList.stream().filter(x -> TradeTypeEnum.SALE.getType().equalsIgnoreCase(x.getType())).map(syncTrade -> syncTrade.getStdAmount().subtract(syncTrade.getFavAmount()))
                .reduce(BigDecimal.ZERO, BigDecimal::add);
        BigDecimal refundTotal = buyGoodList.stream().filter(x -> TradeTypeEnum.REFUND.getType().equalsIgnoreCase(x.getType())).map(syncTrade -> syncTrade.getStdAmount().subtract(syncTrade.getFavAmount()))
                .reduce(BigDecimal.ZERO, BigDecimal::add);
        buyTotal = buyTotal.compareTo(refundTotal) >= 0 ? buyTotal.subtract(refundTotal) : BigDecimal.ZERO;
        return buyTotal;
    }

    /**
     * 获取购买交易总数
     * @param buyGoodList
     * @return
     */
    private BigDecimal getAmountTotal(List<SyncTrade> buyGoodList){
        BigDecimal buyTotal =  buyGoodList.stream().filter(x -> TradeTypeEnum.SALE.getType().equalsIgnoreCase(x.getType())).map(syncTrade->syncTrade.getQty())
                .reduce(BigDecimal.ZERO, BigDecimal::add);
        BigDecimal refundTotal = buyGoodList.stream().filter(x -> TradeTypeEnum.REFUND.getType().equalsIgnoreCase(x.getType())).map(syncTrade->syncTrade.getQty())
                .reduce(BigDecimal.ZERO, BigDecimal::add);
        buyTotal = buyTotal.compareTo(refundTotal) >= 0 ? buyTotal.subtract(refundTotal) : BigDecimal.ZERO;
        return buyTotal;
    }

    @Override
    public List<DrawModel> listMinePrizeList(String phone) {
        ActivityConfig activityConfig = activityConfigService.getConfig();
        if(activityConfig == null){
            throw  new BusinessException(ErrorEnum.SYSTEM_EXP, "未获取活动有效配置");
        }
        List<AccountPrize> prizeList = accountPrizeService.listMinePrize(phone, activityConfig.getActivityId());
        return getDrawModels(prizeList);
    }

    @Override
    public void updateMinePrizeAddress(String phone, Long myPrizeId, String consigneeName, String consigneePhone, String consigneeAddress, String consigneeIdcard) {
        AccountPrize accountPrize = accountPrizeService.getById(myPrizeId);
        // 有奖品 手机号正确 且信息没填
        if (!Objects.isNull(accountPrize) && phone.equals(accountPrize.getAccountPhone()) && PrizeTypeEnum.PHYSICAL.getCode().equals(accountPrize.getPrizeType())) {
            if (StringUtils.isBlank(accountPrize.getConsigneeName()) && StringUtils.isBlank(accountPrize.getConsigneePhone())
                    && StringUtils.isBlank(accountPrize.getConsigneeAddress()) && StringUtils.isBlank(accountPrize.getConsigneeIdcard())) {
                AccountPrize prizeUpdate = new AccountPrize();
                prizeUpdate.setId(myPrizeId);
                prizeUpdate.setConsigneeAddress(consigneeAddress);
                prizeUpdate.setConsigneeIdcard(consigneeIdcard);
                prizeUpdate.setConsigneeName(consigneeName);
                prizeUpdate.setConsigneePhone(consigneePhone);
                // 填写信息后重新同步
                prizeUpdate.setSyncStatus(YesOrNoEnum.NO.getCode());
                accountPrizeService.saveOrUpdate(prizeUpdate);
            } else {
                throw new BusinessException(ErrorEnum.PARAM_CHECK_EXP, "收货信息已填写!");
            }
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public DrawModel drawClick(String phone) throws InterruptedException {
        // 同一用户不可多次重复点击
        RLock lock = redissonClient.getLock(buildKey(phone));
        try {
            lock.tryLock(5, 10, TimeUnit.SECONDS);
            //获取配置
            ActivityConfig activityConfig = activityConfigService.getConfig();
            // 1 验证是否可以抽奖
            LocalDateTime nowTime = LocalDateTime.now();
            LocalDateTime start = nowTime.minusDays(activityConfig.getValidTime()-1);
            start = LocalDateTimeUtil.beginOfDay(start);
            LocalDateTime end = LocalDateTimeUtil.endOfDay(nowTime);
            List<SyncTrade> buyGoodList;
            if(activityConfig.getActivityStoreStatus() == ActivityStoreStatusEnum.NO_CONFINE.getCode() ){
                buyGoodList = syncTradeService.listCanDrawGoods(phone, start, end, null, activityGoodsService.listAllActivityGoods(activityConfig.getActivityId()));

            }else{
                buyGoodList = syncTradeService.listCanDrawGoods(phone, start, end, activityStoreService.listAllActivityStoreId(activityConfig.getActivityId()), activityGoodsService.listAllActivityGoods(activityConfig.getActivityId()));
            }
             int drawCountTotal =0;
            //判断抽奖方式
            //满额抽奖
            if(activityConfig.getRaffle()== RaffleEnum.AMOUNTUNIT.getCode()){
                BigDecimal buyTotal = getAmountTotal(buyGoodList);
                drawCountTotal = buyTotal.divide(new BigDecimal(activityConfig.getAmountUnit()), RoundingMode.DOWN).intValue();
                drawCountTotal = Math.min(activityConfig.getDrawMaxCount(), drawCountTotal);
                if (drawCountTotal <= 0) {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉，你的购物金额不足，已无抽奖次数");
                }
                //足量抽奖
            }else{
                BigDecimal buyTotal = gainBuyTotal(buyGoodList);
                drawCountTotal = buyTotal.divide(new BigDecimal(activityConfig.getCountUnit()), RoundingMode.DOWN).intValue();
                drawCountTotal = Math.min(activityConfig.getDrawMaxCount(), drawCountTotal);
                if (drawCountTotal <= 0) {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉，您购买的活动商品不足，已无抽奖次数");
                }
            }

            Integer alreadyDrawCount = accountPrizeService.gainAlreadyDrawCount(phone, activityConfig.getActivityId(), start, end);
            // 可抽奖次数> 已抽奖次数
            Prize prize;
            AccountPrize accountPrize;
            if (drawCountTotal - alreadyDrawCount > 0) {
                // 获得奖品
                List<Prize> prizeList = gainPrizeInfo(phone,activityConfig);
                // 抽奖
                prize = selectCoupon(prizeList);
                // 乐观锁效验抽到的奖品它合不合规
                if (Integer.valueOf(1).equals(prizeService.updateCount(prize))) {
                    log.info("phone={}的用户抽中了={}", phone, prize);
                } else {
                    // 测试update=0的是否也会回滚
                    prize = prizeList.stream().filter(x -> PrizeTypeEnum.VIRTUAL.getCode().equals(x.getPrizeType())).max(Comparator.comparing(Prize::getId))
                            .orElseThrow(() -> new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "当前抽奖人数过多，请稍候重试!"));
                }
                // 记录抽奖信息
                accountPrize = new AccountPrize();
                accountPrize.setActivityId(activityConfig.getActivityId());
                accountPrize.setPrizeId(prize.getId());
                accountPrize.setAccountName(phone);
                accountPrize.setAccountPhone(phone);
                accountPrize.setPrizeName(prize.getPrizeName());
                accountPrize.setPrizeType(prize.getPrizeType());
                accountPrizeService.saveOrUpdate(accountPrize);
                // 发券
                if (PrizeTypeEnum.VIRTUAL.getCode().equals(prize.getPrizeType()) && StringUtils.isNotBlank(prize.getCouponId())) {
                    otherManager.sendCoupon(prize.getCouponId(), phone, accountPrize.getId());
                }
            } else {
                if (alreadyDrawCount >= activityConfig.getDrawMaxCount()) {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉,你的今日可抽奖次数已用完,请明日再来哦");
                } else {
                    throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "抱歉，你的购物金额或购买活动商品不足，已无抽奖次数");
                }
            }
            // 2 奖品
            BeanCopier<DrawModel> beanCopier = BeanCopier.create(prize, DrawModel.builder().build(), CopyOptions.create());
            DrawModel drawModel = beanCopier.copy();
            drawModel.setMyPrizeId(accountPrize.getId());
            drawModel.setId(prize.getId());
            return drawModel;
        } catch (Exception e) {
            log.error("抽奖异常:{}", e.getMessage());
            throw e;
        } finally {
            lock.unlock();
        }
    }

    public static void main(String[] args) {
        LocalDateTime nowTime = LocalDateTime.now();
        LocalDateTime start = nowTime.minusDays(1-1);
        start = LocalDateTimeUtil.beginOfDay(start);
        LocalDateTime end = LocalDateTimeUtil.endOfDay(nowTime);

    }

    /**
     * 用户限中奖品后重新获取奖品List
     * 1. 获取限制的奖品Id
     * 2. 查询当前号码是否满足中奖限制
     * 3. 如果满足中奖限制，将该奖品剔除
     * @param
     */
    public List<Prize>  gainPrizeInfo(String phone,ActivityConfig activityConfig){
        List<Prize> prizeList = prizeService.listAllPrizes(activityConfig.getActivityId());
        if (CollectionUtils.isEmpty(prizeList)) {
            throw new BusinessException(ErrorEnum.DRAW_CHECK_EXP, "活动已结束!");
        }
        if(activityConfig.getPrizeId() != null){
            int count = accountPrizeService.gainAlreadyDrawCountByPrize(phone,activityConfig.getActivityId(),activityConfig.getPrizeId());
            if(count>=1){
                prizeList= prizeList.stream().filter(p->!activityConfig.getPrizeId().equals(p.getId()) ).collect(Collectors.toList());
            }
            System.out.println(Arrays.toString(prizeList.toArray()));
        }
        return prizeList;
    }



    @Override
    public List<DrawModel> listAllWinningInfo(Integer size) {
        ActivityConfig activityConfig = activityConfigService.getConfig();
        if(activityConfig == null){
            throw  new BusinessException(ErrorEnum.SYSTEM_EXP, "未获取活动有效配置");
        }
        List<AccountPrize> prizeList = accountPrizeService.listAllWinningInfo(size, activityConfig.getActivityId());
        return getDrawModels(prizeList);
    }

    /**
     * 统一封装
     *
     * @param prizeList
     * @return
     */
    private List<DrawModel> getDrawModels(List<AccountPrize> prizeList) {
        List<DrawModel> drawModelList = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(prizeList)) {
            drawModelList = prizeList.stream().map(accountPrize -> {
                BeanCopier<DrawModel> copier = BeanCopier.create(accountPrize, DrawModel.builder().build(), CopyOptions.create());
                return copier.copy();
            }).collect(Collectors.toList());
        }

        return drawModelList;
    }

    /**
     * 抽奖
     *
     * @param prizeList
     * @return
     */
    private Prize selectCoupon(List<Prize> prizeList) {
        TreeMap<Double, Prize> weightMap = new TreeMap<>();
        for (Prize prize : prizeList) {
            //统一转为doubleni
            double lastWeight = weightMap.size() == 0 ? 0 : weightMap.lastKey();
            //权重累加
            weightMap.put(prize.getPrizeOdds().doubleValue() + lastWeight, prize);
        }
        double randomWeight = weightMap.lastKey() * Math.random();
        SortedMap<Double, Prize> tailMap = weightMap.tailMap(randomWeight, false);
        return weightMap.get(tailMap.firstKey());
    }



    /**
     *
     * 加锁可以
     *
     * @param phone
     * @return
     */
    private String buildKey(String phone) {
        StringBuffer sb = new StringBuffer();
        return sb.append(ConStants.SHIZU_DRAW_PREFIX).append(phone).toString().intern();
    }
}
